from typing import Any

from vsview.api import WidgetPluginBase
from vsview.api import hookimpl as _vsview_hookimpl

from .plugin import FramePropsPlugin


@_vsview_hookimpl(tryfirst=True)
def vsview_register_toolpanel() -> type[WidgetPluginBase[Any, Any]]:
    return FramePropsPlugin


@_vsview_hookimpl(tryfirst=True)
def vsview_register_tooldock() -> type[WidgetPluginBase[Any, Any]]:
    return FramePropsPlugin
